# -*- coding: utf-8 -*-

import urllib
import urllib2
import cookielib
import ssl
import traceback
import logging
import MultipartPostHandler

from . import API_USER_AGENT
from logger import get_logger


class CAS(object):
    """ Encapsulate CAS SSO authentication
    """

    # Constants
    __sso_cas_host = 'emc.online.etunexus.com'
    __cas_ticket_endpoint = '/cas/v1/tickets'
    __request_common_headers = {
        'Content-type': 'test/html',
        'Accept': 'application/json, text/plain, */*',
        'Pragma': 'no-cache'
    }

    # Methods
    def __init__(self, cas_group, cas_username, cas_password, cas_host=None, secure=True, loglevel=logging.INFO):
        """ Constructor

        Args:
            cas_group (str): The group to authenticate with CAS
            cas_username (str): The username to authenticate with CAS
            cas_password (str): The password to authenticate with CAS
            cas_host (str): The host/IP of the CAS server
            secure (bool): Enable the certificate check or not
            loglevel (int): Log level
        """
        assert cas_group and len(cas_group) > 0
        self._cas_group = cas_group
        assert cas_username and len(cas_username) > 0
        self._cas_username = cas_username
        assert cas_password and len(cas_password) > 0
        self._cas_password = cas_password

        self._cas_host = self.__sso_cas_host if not cas_host else cas_host
        self._tgt = None
        self._logger = get_logger(loglevel)

        # Init urllib2
        self._init_urllib(secure)

        self._logger.debug("CAS object (%s,%s,%s,%s) constructed" % (cas_host, cas_group, cas_username, cas_password))

    def _init_urllib(self, secure, debuglevel=0):
        cj = cookielib.CookieJar()
        no_proxy_support = urllib2.ProxyHandler({})
        cookie_handler = urllib2.HTTPCookieProcessor(cj)
        ctx = None
        if not secure:
            self._logger.info('[WARNING] Skip certificate verification.')
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        https_handler = urllib2.HTTPSHandler(debuglevel=debuglevel, context=ctx)
        opener = urllib2.build_opener(no_proxy_support,
                                      cookie_handler,
                                      https_handler,
                                      MultipartPostHandler.MultipartPostHandler)
        opener.addheaders = [('User-agent', API_USER_AGENT)]
        urllib2.install_opener(opener)

    @property
    def logger(self):
        """ Get logger """
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    @property
    def cas_host(self):
        """ Get CAS host """
        return self._cas_host

    @cas_host.setter
    def cas_host(self, cas_host):
        self._logger.debug("set cas host: %s" % cas_host)
        self._cas_host = cas_host

    @property
    def cas_group(self):
        """ Get CAS group name """
        return self._cas_group

    @cas_group.setter
    def cas_group(self, cas_group):
        self._logger.debug("set cas group: %s" % cas_group)
        self._cas_group = cas_group

    @property
    def cas_username(self):
        """ Get CAS username """
        return self._cas_username

    @cas_username.setter
    def cas_username(self, cas_username):
        self._logger.debug("set cas usernmae: %s" % cas_username)
        self._cas_username = cas_username

    @property
    def cas_password(self):
        """ Get CAS password """
        return self._cas_password

    @cas_password.setter
    def cas_password(self, cas_password):
        self._logger.debug("set cas password: %s" % cas_password)
        self._cas_password = cas_password

    @property
    def tgt(self):
        """ Get Ticket Granting Ticket (TGT) """
        return self._tgt

    def login(self):
        """ Login CAS

        Grab the Ticket Granting Ticket (TGT)
        Reference : https://wiki.jasig.org/display/casum/restful+api

        Returns:
            str: The tgt granted, or None if failed.
        """
        try:
            if not self._tgt:
                self._tgt = self._request_tgt()
            return self._tgt
        except Exception as e:
            self._logger.error(e.message)
            self._logger.debug(traceback.format_exc())
            return None

    def _request_tgt(self):
        """ Get Ticket Grant Ticket
        """

        ticket_service = 'https://%s%s' % (self._cas_host, self.__cas_ticket_endpoint)
        raw_params = {
            'username': '%s\\%s' % (self._cas_group, self._cas_username),
            'password': self._cas_password
        }
        params = urllib.urlencode(raw_params)
        headers = self.__request_common_headers.copy()
        headers['Content-Length'] = len(params)

        req = urllib2.Request(ticket_service, data=params, headers=headers)
        res = urllib2.urlopen(req)

        if not (res.getcode() / 100) == 2:
            raise Exception("Authentication failed. Please check the host and authentication information.")

        location = res.info().getheader('Location')
        self._logger.debug('cas redirect location: %s' % location)

        tgt = location[location.rfind('/') + 1:]
        self._logger.debug('cas tgt: %s' % tgt)

        return tgt



# -*- coding: utf-8 -*-

import os
import urllib
import urllib2
import json
import codecs
import logging
import traceback

from cas import CAS
from logger import get_logger


class BaseApp(object):
    """ The base class for all etu nexus application API """

    # Constants
    __common_headers = {
        'Content-type': 'application/json',
        'Accept': 'application/json, text/plain, */*',
        'Pragma': 'no-cache'
    }

    __post_forms_headers = {
        'Content-type': 'application/x-www-form-urlencoded',
    }

    def __init__(self, cas, app_name, api_host, api_base, shiro_cas_base):
        """ Constructor

        Args:
            cas (object): A valid CAS instance
            app_name (str): The application name
            api_host (str): The host to make API request to
            api_base (str): The API base of the application
            shiro_cas_base (str): The shiro cas base of the application
        """

        assert cas and isinstance(cas, CAS)
        self._cas = cas
        assert app_name
        self._app_name = app_name
        assert api_host
        self._api_host = api_host
        assert api_base
        self._api_base = api_base
        assert shiro_cas_base
        self._shiro_cas_base = shiro_cas_base

        self._st = None
        self._logger = get_logger()

    @property
    def app_name(self):
        """ Get application name """
        return self._app_name

    @property
    def api_host(self):
        """ Get API Host """
        return self._api_host

    @api_host.setter
    def host(self, api_host):
        self._logger.debug("Set API host (%s)" % api_host)
        self._api_host = api_host

    @property
    def api_base(self):
        """ Get API base"""
        return self._api_base

    @property
    def logger(self):
        """ Get logger """
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    def _resolve_cas_ticket_service(self):
        return 'https://{0}/cas/v1/tickets/{1}'.format(self._cas.cas_host, self._cas.tgt)

    def _resolve_shiro_url(self):
        return 'https://{0}{1}'.format(self._api_host, self._shiro_cas_base)

    def _resolve_shiro_validation_url(self):
        return '{0}?ticket={1}'.format(self._resolve_shiro_url(), self._st)

    def _resolve_api_url(self, api_postfix):
        return 'https://{0}{1}{2}'.format(self._api_host, self._api_base, api_postfix)

    def _request_st(self):
        # Grab a service ticket (ST) for app
        ticket_service = self._resolve_cas_ticket_service()
        raw_params = {
            'service': self._resolve_shiro_url()
        }
        params = urllib.urlencode(raw_params)
        headers = self.__common_headers.copy()
        headers['Content-Length'] = len(params)

        req = urllib2.Request(ticket_service, data=params, headers=headers)
        self._st = urllib2.urlopen(req).read()
        self.logger.debug('cas st (%s) for application (%s)' % (self._st, self._app_name))

        # Login service
        url = self._resolve_shiro_validation_url()
        try:
            urllib2.urlopen(url)
        except urllib2.HTTPError as e:
            if e.getcode() != 403:
                raise e

    def _login_service(self):
        # Login service
        shiro_validation = self._resolve_shiro_validation_url()
        try:
            urllib2.urlopen(shiro_validation)
        except urllib2.HTTPError as e:
            if e.getcode() != 403:
                raise e

    def _check_login(self):
        if not self._st:
            self._logger.error('Please login service before firing API call.')
            raise Exception('Please login service before firing API call.')

    def login(self):
        """ Login service with valid CAS TGT

        Grab the Service Ticket (ST)
        Reference : https://wiki.jasig.org/display/casum/restful+api

        Returns:
            str: The st granted, or None if failed.
        """
        try:
            if not self._st:
                # Ensure login CAS done
                self._cas.login()
                # Make service login
                self._request_st()
                self._login_service()
            return self._st
        except Exception as e:
            self._logger.error(e.message)
            self._logger.debug(traceback.format_exc())
            del self._st
            raise e

    def logout(self):
        """ Logout service with cleaning granted Service Ticket """
        del self._st

    @staticmethod
    def _convert_value(v):
        if isinstance(v, dict):
            return json.dumps(v)
        elif isinstance(v, str):
            return v
        elif isinstance(v, unicode):
            return v.encode('utf-8')
        else:
            return str(v)

    def _request(self, api, data=None, data_serializer=None, file=None, headers=None, method=None):
        if data:
            assert isinstance(data, dict)
        if data_serializer is None:
            data_serializer = json.dumps
        self._check_login()

        # Preparing data and headers if required
        url = self._resolve_api_url(api)

        final_headers = self.__common_headers.copy()
        if headers:
            final_headers.update(headers)

        final_data = None
        if data:
            if file:
                final_data = dict([(k, self._convert_value(v)) for k, v in data.iteritems() if v is not None])
                # [IMPORTANT] Force transform all values to str while MultipartPostHandler append it with str directly.
                # /Library/Python/2.7/site-packages/MultipartPostHandler-0.1.0-py2.7.egg/MultipartPostHandler.pyc in multipart_encode(vars, files, boundary, buffer)
                # 91             buffer += '--%s\r\n' % boundary
                # 92             buffer += 'Content-Disposition: form-data; name="%s"' % key
                # ---> 93             buffer += '\r\n\r\n' + value + '\r\n'
                # 94         for(key, fd) in files:
                # 95             file_size = os.fstat(fd.fileno())[stat.ST_SIZE]
                #
                # TypeError: cannot concatenate 'str' and 'int' objects

                final_data.update({'file': open(file, 'rb')})
                self._logger.debug('Upload file. The content length is calculated before sending request. Not set here.')
            else:
                final_data = data_serializer(data)
                self._logger.debug('Pure post data. Set content length to %d' % len(final_data))
                final_headers['Content-Length'] = len(final_data)
        else:
            self._logger.debug('No data in request. No content length set.')

        # fire api
        req = urllib2.Request(url, data=final_data, headers=final_headers)
        if method:
            req.get_method = lambda: method
        self._logger.debug('%s %s' % (req.get_method(), url))
        self._logger.debug('Request header: %s' % str(req.headers))
        self._logger.debug("Request has body: %s. Request body: %s" %
                            (('Yes' if req.has_data() else 'No'), req.get_data()))

        res = urllib2.urlopen(req).read()
        self._logger.debug('Response: %s' % res)
        try:
            return json.loads(res)
        except ValueError as e:
            self._logger.error('%s error: Illegal response (%s) from server' % (traceback.extract_stack()[-3][2], res))
            raise

    def _download(self, url, save_path):
        self._check_login()

        res = urllib2.urlopen(url)
        with open(save_path, 'wb') as output:
            output.write(res.read())

        return save_path

    def request_get(self, api, headers=None):
        return self._request(api, headers=headers)

    def request_post(self, api, data, headers=None):
        return self._request(api, data=data, headers=headers)

    @staticmethod
    def urlencode_serializer(dict_obj):
        filtered_obj = filter(lambda y: y[1] is not None, dict_obj.iteritems())
        filtered_obj = [(x, codecs.encode(y,'utf-8')) if isinstance(y, unicode) else (x, y) for (x, y) in filtered_obj]
        return urllib.urlencode(filtered_obj)

    def request_post_form(self, api, data, headers=None):
        headers = {} if headers is None else headers.copy()
        headers.update(self.__post_forms_headers)
        return self._request(api, data=data, headers=headers,
                             data_serializer=self.urlencode_serializer)

    def request_post_multipart(self, api, data, headers=None):
        return self._request(api, data=data, file=os.devnull, headers=headers)

    def request_del(self, api, headers=None):
        return self._request(api, headers=headers, method='DELETE')

    def request_upload(self, api, data, file, headers=None):
        return self._request(api, data=data, file=file, headers=headers)

    def request_download(self, url, save_path):
        return self._download(url, save_path)


class BaseAppDict(dict):
    def __init__(self, *args, **kwargs):
        super(BaseAppDict, self).__init__(*args, **kwargs)

    def __getattribute__(self, item):
        if item in self:
            return self[item]
        else:
            return super(BaseAppDict, self).__getattribute__(item)



# -*- coding: utf-8 -*-

from datetime import date

from baseapp import BaseApp
from enum import *
from emc import Group, DataSource, User


class GeneUIInfo(dict):
    """ Structure for gene UI info

    Fields:
        Xaxis_id (str): The string id of gene chart x-axis
        Yaxis_id (str): The string id of gene chart y-axis
        unit (str): The string id of gene unit (used on gene filter UI, etc.)
        Xaxis (str): The (localized) string of gene chart x-axis directly set by system operator
        Xaxis (str): The (localized) string of gene chart y-axis directly set by system operator
    """

    def __init__(self, x_axis_id=None, y_axis_id=None, unit=None, x_axis=None, y_axis=None):
        super(GeneUIInfo, self).__init__({
            'Xaxis_id': x_axis_id,
            'Yaxis_id': y_axis_id,
            'unit': unit,
            'Xaxis': x_axis,
            'Yaxis': y_axis
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj.get('Xaxis_id'), dict_obj.get('Yaxis_id'), dict_obj.get('unit'),
                   dict_obj.get('Xaxis'), dict_obj.get('Yaxis'))


class GeneCategory(dict):
    """ Structure for (aggregation) gene category.

    Fields:
        name (str): Fixed gene name.
        subcategories (list): A list of GeneCategory as subcategories.
        genes (list): A list of Gene instances.

        id (int): The auto id.
    """

    def __init__(self, name, subcategories, genes, id=None):
        assert name and subcategories is not None and isinstance(subcategories,
                                                                 list) and genes is not None and isinstance(genes, list)
        super(GeneCategory, self).__init__({
            'name': name,
            'subcategories': [x if isinstance(x, GeneCategory) else GeneCategory.from_dict(x) for x in
                              subcategories],
            'genes': [x if isinstance(x, Gene) else Gene.from_dict(x) for x in genes],
            'id': id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['subcategories'], dict_obj['genes'], dict_obj.get('id'))


class Gene(dict):
    """ Structure for (aggregation) gene.

    Fields:
        id (str): The id of the gene. It should be unique in a subcategory.
        name (str); Gene name.
        timerange (int): Gene time range.
        type (str): Gene type, refer to "GeneType" for valid values.
        chartType (str): Gene chart type, refer to "GeneChartType" for valid values.
        uiInfo (obj): A GeneUIInfo instance as the gene UI info setting.
    """

    def __init__(self, id, name, time_range, type, chart_type, ui_info):
        assert id and name and time_range is not None and type and chart_type and ui_info
        super(Gene, self).__init__({
            'id': id,
            'name': name,
            'timerange': time_range,
            'type': type,
            'chartType': chart_type,
            'uiInfo': ui_info if isinstance(ui_info, GeneUIInfo) else GeneUIInfo.from_dict(ui_info)
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['id'], dict_obj['name'], dict_obj['timerange'], dict_obj['type'], dict_obj['chartType'],
                   dict_obj['uiInfo'])


class BandCategory(dict):
    """ Structure for a band category.

    Fields:
        name (str): The band category name.
        bands (list): A list of Band instances.

        id (int): The auto id
    """

    def __init__(self, name, bands=None, id=None):
        assert name
        if bands is None:
            bands = []
        assert isinstance(bands, list)
        super(BandCategory, self).__init__({
            'name': name,
            'bands': [x if isinstance(x, Band) else Band.from_dict_with_ext_cat(x, id) for x in bands],
            'id': id
        })

    def to_simple(self):
        ret = {'name': self['name']}
        if self.get('isDefault'):
            ret['isDefault'] = self.get('isDefault')
        return ret

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj.get('bands'), dict_obj.get('id'))


class BandGene(dict):
    """ Structure for a gene-based band.

    Fields:
        geneId (str): The gene id.
        cid (int): The data source id to calculate the band.
        operator (str): The band operator to calculate the band, refer to "BandGeneOperator" enum for valid values.
        operand (str): The operands to calculate the band.
    """

    def __init__(self, gene_id, data_source, operator, operand):
        assert gene_id and data_source >= 0 and operator and operand is not None
        super(BandGene, self).__init__({
            'geneId': gene_id,
            'cid': data_source['id'] if isinstance(data_source, DataSource) else int(data_source),
            'operator': operator,
            'operand': operand
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['geneId'], dict_obj['cid'], dict_obj['operator'], dict_obj['operand'])


class BandCombine(dict):
    """ Structure for a combined band.

    Fields:
        bandIds (list): List of band ids to combine.
        operators (list): List of operators used to combine the bands (only the first element is used so far), refer to
        "BandCombineOperator" enum for valid values.
    """

    def __init__(self, bands, operators):
        assert bands and isinstance(bands, list)
        assert isinstance(operators, list)
        super(BandCombine, self).__init__({
            'bandIds': [x['id'] if isinstance(x, Band) else x for x in bands],
            'operators': operators
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['bandIds'], dict_obj['operators'])


class Band(dict):
    """ Structure for a band.
    
    Fields:
        categoryId (int): The band category id of the band belongs to.
        name (str): Band name.
        description (str): Band description (not used so far).
        type (str): Band type, refer to 'BandType' for valid values.
        targetGene (obj): A BandGene instance if type=gene or type=fixedgene.
        targetBand (obj): A BandCombine instance if type=combine.

        needRefresh (bool): Need to refresh the band or not.
        snapshotInfo (obj): A SnapshotInfo instance. For a pure band, it's an SnapshopInfo with no parent and empty
        children.

        shared (bool): The band is shared in group or not.

        id (int): The auto id.
        amount (int): Amount of users in the band.
        updateTime (long): The band update time in Epoch (milliseconds).
        owner (str): The owner (user_id) of the band.
    """

    def __init__(self, category, name, description,
                 type, target_gene=None, target_band=None,
                 need_refresh=True, snapshot_info=None,
                 shared=False,
                 id=None, amount=None, update_time=None, owner=None):
        super(Band, self).__init__({
            'categoryId': category['id'] if isinstance(category, BandCategory) else int(category),
            'name': name,
            'description': description,
            'needRefresh': need_refresh,
            'type': type,
            'targetGene': target_gene if isinstance(target_gene, BandGene)
                else BandGene.from_dict(target_gene) if target_gene is not None else None,
            'targetBand': target_band if isinstance(target_band, BandCombine)
                else BandCombine.from_dict(target_band) if target_band is not None else None,
            'snapshotInfo': snapshot_info if isinstance(snapshot_info, SnapshotInfo)
                else SnapshotInfo.from_dict(snapshot_info) if snapshot_info is not None else SnapshotInfo(),
            'shared': shared,

            'id': id,
            'amount': amount,
            'updateTime': update_time,
            'owner': owner
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['categoryId'], dict_obj['name'], dict_obj['description'],
                   dict_obj['type'], dict_obj.get('targetGene'), dict_obj.get('targetBand'),
                   dict_obj['needRefresh'], dict_obj.get('snapshotInfo'),
                   dict_obj.get('shared'),
                   dict_obj.get('id'), dict_obj.get('amount'), dict_obj.get('updateTime'),
                   dict_obj.get('owner'))

    @classmethod
    def from_dict_with_ext_cat(cls, dict_obj, category):
        assert dict_obj and category
        category_id = category['id'] if isinstance(category, BandCategory) else int(category)
        return cls(category_id, dict_obj['name'], dict_obj['description'],
                   dict_obj['type'], dict_obj.get('targetGene'), dict_obj.get('targetBand'),
                   dict_obj['needRefresh'], dict_obj.get('snapshotInfo'),
                   dict_obj.get('shared'),
                   dict_obj.get('id'), dict_obj.get('amount'), dict_obj.get('updateTime'),
                   dict_obj.get('owner'))


class SnapshotInfo(dict):
    """ Structure for band snapshot information.

    Fields:
        parent (int): The parent band id.
        children (list): A list of band ids as children.
    """

    def __init__(self, parent=None, children=None):
        if not children:
            children = []
        super(SnapshotInfo, self).__init__({
            'parent': parent,
            'children': children
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj.get('parent'), dict_obj.get('children'))


class FixedGeneCategory(dict):
    """ Structure for fixed (external) gene categories

    Fields:
        name (str): Fixed gene name
        subcategories (list): A list of FixedGeneCategory as subcategories
        genes (list): A list of

        id (int): The auto id
    """

    def __init__(self, name, subcategories, genes, id=None):
        assert name and subcategories is not None and isinstance(subcategories,
                                                                 list) and genes is not None and isinstance(genes, list)
        super(FixedGeneCategory, self).__init__({
            'name': name,
            'subcategories': [x if isinstance(x, FixedGeneCategory) else FixedGeneCategory.from_dict(x) for x in
                              subcategories],
            'genes': [x if isinstance(x, FixedGene) else FixedGene.from_dict(x) for x in genes],
            'id': id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['subcategories'], dict_obj['genes'], dict_obj.get('id'))


class FixedGene(dict):
    """ Structure for fixed (external) gene.

    Fields:
        id (str): The id of the fixed gene. It should be unique in a fixed gene subcategory.
        name (str); Gene name.
        timerange (int): Gene time range.
        type (str): Gene type, refer to "GeneType" enum for valid values.
        chartType (str): Gene chart type, refer to "GeneChartType" for valid values.
        uiInfo (obj): A GeneUIInfo instance as the gene UI info setting.
    """

    def __init__(self, id, name, time_range, type, chart_type, ui_info):
        assert id and name and time_range is not None and type and chart_type and ui_info
        super(FixedGene, self).__init__({
            'id': id,
            'name': name,
            'timerange': time_range,
            'type': type,
            'chartType': chart_type,
            'uiInfo': ui_info if isinstance(ui_info, GeneUIInfo) else GeneUIInfo.from_dict(ui_info)
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['id'], dict_obj['name'], dict_obj['timerange'], dict_obj['type'], dict_obj['chartType'],
                   dict_obj['uiInfo'])


class PopulationSummary(dict):
    """ Structure for population summary item.

    Fields:
        key (str): The key value of the summary item (depending on the gene).
        amount (int): The segment size with the key value.
    """
    def __init__(self, key, amount):
        super(PopulationSummary, self).__init__({
            'key': key,
            'amount': amount
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['key'], dict_obj['amount'])


class PopulationTimeline(dict):
    """ Structure for poluation timeline data.
    
    Fields:
        name (str): The gene id.
        group (str): The group name.
        data_source_id (int): The data source id.
        start_time (int): The Epoch timestamp of data query start time.
        end_time (int): The Epoch timestamp of data query end time.
        data (obj): List of (date, count) detail data.
    """

    def __init__(self, name, group, ds_id, start_time, end_time, data):
        super(PopulationTimeline, self).__init__({
            'name': name,
            'group': group,
            'data_source_id': ds_id,
            'start_time': start_time,
            'end_time': end_time,
            'data': [(date.fromtimestamp(x[0]/1000), x[1]) for x in data]
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['group'], dict_obj['cid'], dict_obj['startTime'], dict_obj['endTime'],
                   dict_obj['data'])


class EIGroup(dict):
    """ Structure for Etu Insight group.

    Fields:
        name (str): The group id/name.
        displayName (str): The display name.

        id (int): The auto id (the same id as the same group in EMC).
    """

    def __init__(self, name, display_name, id=None):
        super(EIGroup, self).__init__({
            'name': name,
            'displayName': display_name,
            'id': id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj.get('id'))

    @classmethod
    def from_emc_group(cls, emc_group):
        assert emc_group and isinstance(emc_group, Group)
        return cls(emc_group['name'], emc_group['displayName'], emc_group['id'])


class EIUser(dict):
    """ Structure for Etu Insight user.

    Fields:
        name (str): The login name of the user.
        displayName (str): The display name.
        role (str): The user role in Etu Insight, refer to "AppRoleName" for valid values.
                    It's a single role name rather than a list as the roles in emc.User object for multiple applications.
        group (object): A emc.Group or EIGroup instance that the user belongs to.

        id (int): The auto id (the same id as the same user in EMC)
    """

    def __init__(self, name, display_name, role, group, id=None):
        super(EIUser, self).__init__({
            'name': name,
            'displayName': display_name,
            'role': role,
            'group': group if isinstance(group, EIGroup) else
                     EIGroup.from_emc_group(group) if isinstance(group, Group) else EIGroup.from_dict(group),
            'id': id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj['role'], dict_obj['group'], dict_obj.get('id'))

    @classmethod
    def from_emc_user(cls, emc_group, emc_user):
        assert emc_group
        assert emc_user and isinstance(emc_user, User)
        filtered_role = filter(lambda u: u['appId'] == AppId.EI, emc_user.roles)
        if len(filtered_role) == 0:
            raise RuntimeError('The user (%s) does not have EI authorization.' % emc_user['name'])

        ei_role = filtered_role[0]
        return cls(emc_user['name'], emc_user['displayName'], ei_role['roleName'], emc_group, emc_user['id'])


class UidBandList(dict):
    """ Structure for the band list (id and name) of a uid belongs to.

    Fields:
        userId (str): The uid/customer id.
        bandIdList (list): The list of band ids.
        bandNameList (list): The list of band names.
    """

    def __init__(self, band_id_list, band_name_list):
        assert band_id_list and isinstance(band_id_list, list)
        assert band_name_list and isinstance(band_name_list, list)
        super(UidBandList, self).__init__({
            'bandIdList': band_id_list,
            'bandNameList': band_name_list
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['bandIdList'], dict_obj['bandNameList'])


class EI3(BaseApp):
    """ Encapsulate Etu Insight (v3) API """

    __APP_NAME = 'EI3'
    __HOST = 'eihome.online.etunexus.com'
    __API_BASE = '/EI/V3/service/v1'
    __SHIRO_CAS_BASE = '/EI/V3/shiro-cas'

    def __init__(self, cas, host=None, api_base=None, shiro_cas_base=None):
        """ Constructor of EI3 instance. """

        api_base = api_base if api_base else self.__API_BASE
        super(EI3, self).__init__(cas, EI3.__APP_NAME,
                                  api_host=host if host else self.__HOST,
                                  api_base=api_base if api_base else self.__API_BASE,
                                  shiro_cas_base=shiro_cas_base if shiro_cas_base else self.__SHIRO_CAS_BASE)

    def _resolve_root_url(self, postfix):
        return 'https://{0}/{1}'.format(self._api_host, postfix)

    # User info
    def get_me(self):
        """ Get the special "me" user information.

        The "me" might be a simulated user if a "su" process is applied by calling do_su_login().

        Arguments:
        Return:
            A EIUser instance as "me".
        """
        res = self.request_get('/user/me')
        return EIUser.from_dict(res['data'])

    def get_superme(self):
        """ Get the special "superme" user information.

        The "superme" is the actual login user no matter a "su" process is applied or not.

        Arguments:
        Return:
            A EIUser instance as "superme".
        """
        res = self.request_get('/user/superme')
        return EIUser.from_dict(res['data'])

    # Gene category and detail gene info
    def get_gene_categories(self, group):
        """ Get all (aggregation) gene categories and detail genes.

        Though the gene setting is global, different group may be authorized with different set.

        Arguments:
            group (obj or int): The emc.Group instance, or EIGroup instance, or a group id to get genes.
        Return:
            A list of GeneCategory instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
        res = self.request_get('/genecategory?groupId={0}'.format(group_id))
        return [GeneCategory.from_dict(x) for x in res['data']]

    # Band category
    def get_band_categories(self):
        """ Get user's all band categories and detail bands.

        Arguments:
        Return:
            A list of BandCategory instances.
        """
        res = self.request_get('/bandcategory')
        return [BandCategory.from_dict(x) for x in res['data']]

    def add_band_category(self, band_category):
        """ Add a new band category.

        In most cases, no need to initialize the "bands" in the new category. Use add_band() to add bands into the new
        category later.

        Arguments:
            band_category (obj): The BandCategory instance to add.
        Return:
            A BandCategory instance as the added one (with fields filled by server, e.g. id).
        """
        assert band_category and isinstance(band_category, BandCategory)
        res = self.request_post_form('/bandcategory', band_category.to_simple())
        return BandCategory.from_dict(res['data'])

    def update_band_category(self, band_category):
        """ Update a band category.

        Only update the attributes of a band category, e.g. "name". DO NOT USE this method to manipulate "bands". Use
        add_band(), update_band(), delete_band() methods instead.

        Arguments:
            band_category (obj): The BandCategory instance to update.
        Return:
            A BandCategory instance as the updated one.
        """
        assert band_category and isinstance(band_category, BandCategory)
        band_category_id = band_category['id']
        assert band_category_id
        res = self.request_post_form('/bandcategory/{0}'.format(band_category_id), band_category.to_simple())
        return BandCategory.from_dict(res['data'])

    def del_band_category(self, band_category):
        """ Delete a band category.

        It would cause error if the band category is not empty (still exist bands in it).

        Arguments:
            band_category (obj or int): The BandCategory instance or a band category id to delete.
        Return:
            The band category id deleted. It must be the same as the one in argument.
        """
        assert band_category
        band_category_id = band_category['id'] if isinstance(band_category, BandCategory) else int(band_category)
        assert band_category_id
        res = self.request_del('/bandcategory/{0}'.format(band_category_id))
        # For unknown reason, delete category returns a list. Leave a possible change in the future that it might become
        # a simple dict
        res_id = res['data'][0]['id'] if isinstance(res['data'], list) else res['data']['id']
        assert res_id == band_category_id
        return res_id

    def get_shared_band_categories(self, group=None):
        """ Get shared band categories and detail bands in the group.

        Arguments:
            group (obj or int): The emc.Group instance, or EIGroup instance, or a group id to get shared bands. If
                                it is omitted (=None), current user gorup is used. 
        Return:
            A list of BandCategory instances.
        """
        api = '/sharingbandcategory'
        if group is not None:
            group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
            api = '/sharingbandcategory?groupId={0}'.format(group_id)

        res = self.request_get(api)
        return [BandCategory.from_dict(x) for x in res['data']]

    # Band
    def add_band(self, band, file_path=None):
        """ Add a new band.

        If the band type is BandType.UPLOAD, it requires file as the band list. The file format is standard CSV with at
        lease a "uid" column as the user/customer id list.

        Arguments:
            band (obj): The Band instance to add.
            file_path (str): The file to upload if the band type is BandType.UPLOAD.
        Return:
            A Band instance as the added one.
        """
        assert band and isinstance(band, Band)
        band_category_id = band['categoryId']
        if file_path:
            assert band['type'] == BandType.UPLOAD
            res = self.request_upload('/band', band, file_path)
        else:
            res = self.request_post_multipart('/band', band)
        return Band.from_dict_with_ext_cat(res['data'], band_category_id)

    def update_band(self, band, file_path=None):
        """ Update a band.

        It is not recommended to change the band type after added with this method, but not enforced. Therefore, it is
        suggested that client code could make the check before calling update_band().

        If the band type is BandType.UPLOAD, it requires file as the band list. The file format is standard CSV with at
        lease a "uid" column as the user/customer id list.

        Arguments:
            band (obj): The Band instance to add.
            file_path (str): The file to upload if the band type is BandType.UPLOAD.
        Return:
            A Band instance as the updated one.
        """
        assert band and isinstance(band, Band)
        band_category_id = band['categoryId']
        band_id = band['id']
        assert band_id
        if file_path:
            assert band['type'] == BandType.UPLOAD
            res = self.request_upload('/band/{0}'.format(band_id), band, file_path)
        else:
            res = self.request_post_multipart('/band/{0}'.format(band_id), band)
        return Band.from_dict_with_ext_cat(res['data'], band_category_id)

    def del_band(self, band):
        """ Delete a band.

        Arguments:
            band (obj or int): The Band instance or a band id to delete.
        Return:
            The band id deleted. It must be the same as the one in argument.
        """
        assert band
        band_id = band['id'] if isinstance(band, Band) else int(band)
        assert band_id
        res = self.request_del('/band/{0}'.format(band_id))
        res_id = res['data']['id']
        assert res_id == band_id
        return res_id

    def get_uid_list(self, band, save_path):
        """ Get/download user/customer id list in a band.
        
        Arguments:
            band (obj or int): The Band instance or a band id to get the user list.
            save_path (str): The file path to save the user list.
        Return:
            The save_path in argument.
        """
        assert band
        band_id = band['id'] if isinstance(band, Band) else int(band)
        assert band_id
        res = self.request_get('/band/{0}/uidlist'.format(band_id))
        download_link = res['data']['downloadLink']
        self.logger.debug('Get the download link: %s' % download_link)
        download_url = self._resolve_root_url(download_link)
        self.logger.debug('Make download from: %s' % download_url)
        return self.request_download(download_url, save_path)

    # Snapshot
    def do_snapshot(self, band):
        """ Make a snapshot/clone of a band.

        Arguments:
            band (obj or int): The Band instance or a band id to clone.
        Return:
            A Band instance as the cloned one, with a given name derived from the original band name.
        """
        assert band
        band_id = band['id'] if isinstance(band, Band) else int(band)
        assert band_id
        res = self.request_post('/band/{0}/snapshot'.format(band_id))
        return Band.from_dict(res['data'])

    # Fixed (external) gene
    def get_fixed_gene_categories(self):
        """ Get the fixed (external) gene categories and detail genes.

        Arguments:
        Return:
            A list of FixedGeneCategory instances.
        """
        res = self.request_get('/fixedgenecategory')
        return [FixedGeneCategory.from_dict(x) for x in res['data']]

    def upload_fixed_gene_schema(self, group, file_path):
        """ Upload fixed (external) gene schema (admin/operator only).

        It requires a gene schema file, and the file format is JSON. Please refer to the product documents or consult
        with Etu members for the detail.

        Arguments:
            group (obj or int): The emc.Group or EIGroup instance or a group id to upload fixed gene schema.
            file_path (str): The schema file.
        Return:
            A FixedGeneCategory instance as the schema defined.
        """
        assert group and file_path
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
        assert group_id
        res = self.request_upload('/fixedgene/schema', {'groupId': group_id}, file_path)
        return FixedGeneCategory.from_dict(res['data'])

    def upload_fixed_gene_data(self, group, file_path):
        """ Upload fixed (external) gene data (admin/operator only).

        It requires a gene data/value file, and the file format is standard CSV. A "uid" column is required, and other
        columns should be the same as the "gene_id" in the schema uploaded. Please refer to the product documents or
        consult with Etu members for the detail.

        Arguments:
            group (obj or int): The emc.Group or EIGroup instance or a group id to upload fixed gene schema.
            file_path (str): The data file.
        Return:
            A success message.
        """
        assert group and file_path
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
        assert group_id
        res = self.request_upload('/fixedgene/data', {'groupId': group_id}, file_path)
        return res['data']

    # Population summary data
    def get_population_summary(self, gene, data_source):
        """ Get population summary data.

        Arguments:
            gene (obj or int): The Gene instance or a gene id to get summary data.
            data_source (obj or int): The emc.DataSource instance or a data source id.
        Return:
            A list of PopulationSummary instances.
        """
        assert gene and data_source
        gene_id = gene['id'] if isinstance(gene, Gene) else str(gene)
        data_source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        res = self.request_get('/population/summary?geneId={0}&cId={1}'.format(gene_id, data_source_id))
        return [PopulationSummary.from_dict(x) for x in res['data']]

    def get_population_timeline(self, genes, data_source, start_date, end_date, filter_op,
                                filter_comp=EIPopulationTimelineFilterCompType.STRING + ':*'):
        """ Get population timeline data.
        
        Arguments:
            genes (obj): A list of Gene instances or gene ids to get timeline data.
            data_source (obj or int): The emc.DataSource instance or a data source id.
            start_date (date or str): Start date of the data to query. It could be a datetime.date instance, or a
                string in 'yyyy-mm-dd' format.
            end_date (date or str): End date of the data to query. It could be a datetime.date instance, or a
                string in 'yyyy-mm-dd' format.
            filter_op (str): The filter operator, refer to "EIPopulationTimelineFilterOP" enum for valid values.
            filter_comp (str): The filter comparator in format "TYPE:VALUE". For valid "TYPE", refer to
                "EIPopulationTimelineFilterCompType" enum.
        Return:
            A list of PopulationTimeline instances.
        """
        assert genes and isinstance(genes, list) and data_source
        assert start_date and end_date
        assert filter_op and filter_comp
        gene_ids = ','.join([gene['id'] if isinstance(gene, Gene) else str(gene) for gene in genes])
        data_source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        start_date_str = '%04d-%02d-%02d' % (start_date.year, start_date.month, start_date.day) \
            if isinstance(start_date, date) else str(start_date)
        end_date_str = '%04d-%02d-%02d' % (end_date.year, end_date.month, end_date.day) \
            if isinstance(end_date, date) else str(end_date)

        res = self.request_get('/timeline/population/{0}/{1}?startTime={2}&endTime={3}&filterOperator={4}&filterComparator={5}'.format(
            gene_ids, data_source_id, start_date_str, end_date_str, filter_op, filter_comp
        ))
        return [PopulationTimeline.from_dict(x) for x in res['data']]

    # Statistics
    def get_statistics(self, group, items, start_date, end_date=date.today()):
        """ Get EI statistics.

        Arguments:
            group (obj or int): The emc.Group instance, or EIGroup instance, or a group id to get statistics.
            items (list): Refer to EIStatisticsItem for valid values, and put the (multiple) items into a list.
                e.g. [EIStatisticsItem.DOWNLOAD_UID_LIST, EIStatisticsItem.ADD_BAND]
            start_date (date or str): Start date of the statistics to query. It could be a datetime.date instance, or a
                string in 'yyyy-mm-dd' format.
            end_date (date or str): End date of the statistics to query. It could be a datetime.date instance, or a
                string in 'yyyy-mm-dd' format.
        Return:
            A list of tuple:
             [0] A date string in yyyy-mm-dd
             [1] An item:count dictionary

            The result is sorted by the date.
        """
        assert group and items and isinstance(items, list) and start_date and end_date
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
        start_date_str = '%04d-%02d-%02d' % (start_date.year, start_date.month, start_date.day) \
            if isinstance(start_date, date) else str(start_date)
        end_date_str = '%04d-%02d-%02d' % (end_date.year, end_date.month, end_date.day) \
            if isinstance(end_date, date) else str(end_date)
        item_str = ','.join(items)
        res = self.request_get('/statistics/query?groupId={0}&startDate={1}&endDate={2}&itemName={3}'.format(
            group_id, start_date_str, end_date_str, item_str))
        return sorted([(key, value) for (key, value) in res['data'].iteritems()])

    # Customer information
    def get_customer_info(self, data_source, uid):
        """ Get the gene values of a user/customer.

        Arguments:
            data_source (obj or int): A emc.DataSource instance or a data source id.
            uid (str): The user/customer id.
        Return:
            A dict with the key as gene ids and corresponding values.
        """
        assert data_source and uid and isinstance(uid, str)
        data_source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        res = self.request_get('/customerinformation/{0}?cId={1}'.format(uid, data_source_id))
        return res['data']

    def get_uid_band_list(self, user, uid):
        """ Get the band list of a user/customer belongs to.

        This methods returns the band list, owned by the "user" given, of the user/customer belongs to. Currently, there
        is no authentication check.

        Arguments:
            user (obj or int): The owner of the bands.
            uid (str): The user/customer id.
        Return:
            An UidBandList instance.
        """
        assert user and uid and isinstance(uid, str)
        user_id = user['id'] if isinstance(user, User) or isinstance(user, EIUser) else int(user)
        res = self.request_get('/uidbandlist/id?uid={0}&userId={1}'.format(uid, user_id))
        return UidBandList.from_dict(res['data'])

    # Item data source
    def upload_item_data(self, group, file_path):
        """ Upload item information data.

        It requires a item information file as standard CSV format. Please refer to the product documents or consult
        with Etu members for the detail.

        Arguments:
            group (obj or int): The emc.Group instance, or EIGroup instance, or a group id to add the data.
            file_path (str): The data file.
        Return:
            The total item info (pid count) uploaded.
        """
        assert group and file_path
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, EIGroup) else int(group)
        res = self.request_upload('/item', data={'groupId': group_id}, file=file_path)
        return res['data']['pidNumber']

    # EC Application - Summary
    def get_summary(self, summary_id, data_source, band):
        """ Get the summary data of EC application.

        Arguments:
            summary_id (str): The summary id, e.g. Summary_Revisit_7, Summary_Revisit_30, etc.
            data_source (obj or int): A emc.DataSource instance or a data source id.
            band (obj or int): A Band instance of a band id.
        Return:

        """
        assert summary_id and data_source and band
        upper_bound = 1
        lower_bound = 1
        data_source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        band_id = band['id'] if isinstance(band, Band) else int(band)
        res = self.request_get('/summary/{0}?cId={1}&bandId={2}&upperBound={3}&lowerBound={4}'.format(
            summary_id, data_source_id, band_id, upper_bound, lower_bound))
        return res['data']

    # Un-documented APIs for admin/operator
    def do_su_login(self, group, user):
        """ Make "su" login (admin/operator only).

        Arguments:
            group (obj or int): The Group instance, or EIGroup instance, or a group name of the user to simulate.
            user (obj or int): The User instance, or EIUser instance, or a user name of the user to simulate.
        Return:
            A success message.
        """
        assert group and user
        group_name = group['name'] if isinstance(group, Group) or isinstance(group, EIGroup) else str(group)
        user_name = user['name'] if isinstance(user, User) or isinstance(user, EIUser) else str(user)
        res = self.request_get('/suauth?groupName={0}&userName={1}'.format(group_name, user_name))
        return res['data']

    def do_su_logout(self):
        """ Make "su" logout (admin/operator only).

        Arguments:
        Return:
            A success message.
        """
        res = self.request_del('/suauth')
        return res['data']

    def get_default_bandcategories(self, data_source):
        """ Get default band category and bands.
        
        Arguments:
            data_source (obj or int): The emc.DatraSource instance or a data source id.
        Return:
            A list of BandCategory instances.
        """
        assert data_source
        data_source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        res = self.request_get('/defaultbandcategory?cId={0}'.format(data_source_id))
        return [BandCategory.from_dict(x) for x in res['data']]


# -*- coding: utf-8 -*-

API_VERSION = 'v3'
API_USER_AGENT = 'Etu Nexus API Python Client' + '-' + API_VERSION

__version__ = '0.6.7'
__url__ = 'http://www.etunexus.com/'
__author_email__ = 'indichen@etusolution.com'
__author__ = 'Etu Corporation'
__license__ = 'GPL'


# -*- coding: utf-8 -*-


class AppId(object):
    """ Enum of app ids """
    EMC = 'ETU-SHIELD'
    ER = 'ETU-RECOMMENDER'
    EI = 'ETU-INSIGHT'


class AppRoleName(object):
    """ Enum of roles of apps """
    VIEWER = 'Viewer'
    OPERATOR = 'Operator'
    ADMIN = 'ADMIN'


class DataSourceContentType(object):
    """ Enum of data source content type """
    BEHAVIOR = 'behavior'
    ITEM_INFO = 'item'
    USER_PROFILE = 'userprofile'
    USER_RANK = 'userrank'


class DataSourceType(object):
    """ Enumeration of data source type """
    EVENT_COLLECTOR = 'EVENT_COLLECTOR'
    IMPORTER = 'IMPORTER'
    FETCH = 'FETCH'
    UPLOAD = 'UPLOAD'


class LogicAlgType(object):
    """ Enumeration of recommendation logic type """
    USER_BASE = 'USER_BASE'
    ITEM_BASE = 'ITEM_BASE'
    RANK = 'RANK'
    KEYWORD = 'KEYWORD'


class LogicAvlItemFilterMode(object):
    """ Enumeration of recommendation logic item filter mode """
    DISABLED = 'DISABLED'
    INCLUDE = 'INCLUDE'
    EXCLUDE = 'EXCLUDE'


class LogicAlgorithmId(object):
    """ Enumeration of recommendation algorithm samples """
    UNKNOWN = 'UNKNOWN'
    USER_BASED_CF = 'USER_BASED_CF'
    ITEM_BASED_CF = 'ITEM_BASED_CF'
    RANKING = 'RANKING'
    RANKING_ITEMINFO = 'RANKING_ITEMINFO'
    SEARCH2CLICK = 'SEARCH2CLICK'
    INFO_INTEGRITY = 'Info_Integrity'
    ALS = 'ALS'
    LDA = 'LDA'


class LogicExecuteState(object):
    """ Enumeration of recommendation logic execution state """
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'
    RUNNING = 'RUNNING'


class EventAction(object):
    """ Enumeration of action samples used in algorithms """
    VIEW = 'view'
    CART = 'cart'
    ORDER = 'order'
    LIKE = 'like'
    SHARE = 'share'


class BandType(object):
    """ Enumeration of band types """
    GENE = "gene"
    COMBINE = "combine"
    UPLOAD = "upload"
    FIXED_GENE = "fixedgene"


class BandGeneOperator(object):
    """ Enumeration of operators for gene-based bands """
    LT = 'LT'
    LE = 'LE'
    EQ = 'EQ'
    GE = 'GE'
    GT = 'GT'
    IN = 'IN'
    BETWEEN = 'BETWEEN'
    # for list
    INC = 'INC'
    INCALL = 'INCALL'
    INCNONE = 'INCNONE'
    # for range list
    RINC = 'RINC'
    RINCALL = 'RINCALL'
    RINCNONE = 'RINCNONE'
    # partial comparison for listKINC = 'KINC'
    KINCALL = 'KINCALL'
    KINCNONE = 'KINCNONE'


class BandCombineOperator(object):
    """ Enumeration of operators for combined bands """
    UNION = 'UNION'
    INTERSECT = 'INTERSECT'
    EXCEPT = 'EXCEPT'


class GeneType(object):
    """ Enumeration of gene types. It is also for fixed (external) gene """
    SET = 'set'
    DATE = 'date'
    PERCENT = 'percent'
    HOURS = 'hours'
    SET_KEYWORD = 'set_keyword'
    NUMBER_TIME = 'number_time'
    NUMBER = 'number'
    WEEKDAYS = 'weekdays'
    RANGE_DATE = 'range_date'


class GeneChartType(object):
    """ Enumeration of gene chart types. It is also for fixed (external) gene """
    SEMIDONUT = 'semidonut'
    BAR = 'bar'
    LINE = 'line'
    COLUMN = 'column'
    HEATMAP_DATE = 'heatmap_date'


class EIStatisticsItem(object):
    """ Enumeration of EI statistics item name """
    DOWNLOAD_UID_LIST = 'downloadUidList'
    ADD_BAND = 'addBand'
    UPDATE_BAND = 'updateBand'
    BAND_COUNT = 'bandCount'


class EIPopulationTimelineFilterOP(object):
    """ Enumeration of EI population timeline filter op """
    LT = 'LT'
    LE = 'LE'
    EQ = 'EQ'
    GE = 'GE'
    GT = 'GT'
    NE = 'NE'


class EIPopulationTimelineFilterCompType(object):
    """ Enumeration of EI population timeline filter comparator type """
    NUMBER = 'NUMBER'
    STRING = 'STRING'
    SUBSTRING = 'SUBSTRING'


class LayoutTitleAlignment(object):
    """ Enumeration of ER logic layout title alignment """
    CENTER = 'center'
    LEFT = 'left'
    RIGHT = 'right'


class AlgTrainingState(object):
    """ Enumeration of ER algorithm training state """
    NA = 'N/A'
    INIT = 'INIT'
    RUNNING = 'RUNNING'
    DONE = 'DONE'
    FAILED = 'FAILED'



# -*- coding: utf-8 -*-

import time
import json

from baseapp import BaseApp
from enum import *


class AppInfo(dict):
    """ Structure for an app information

    Fields:
        id (str): App id, refer to "AppId" enum for valid values.
        version (str): App version.
        console (str): The URL link to the app console.
        image (str): The URL link to the app presentation image.
    """

    def __init__(self, id, version, console=None, image=None):
        super(AppInfo, self).__init__({
            'id': id,
            'version': version,
            'console': console,
            'image': image
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['id'], dict_obj['version'], dict_obj.get('console'), dict_obj.get('image'))


class AppPermission(dict):
    """ Structure of an app permission.

    Fields:
        category (str): Permission category, refer to the valid values of each apps.
        id (str): Permission id, refer to each apps.
        level (int): FORBIDDEN (0), VIEW (1), SAVE (2), DELETE (3)
    """
    def __init__(self, category, id, level):
        assert category and id and level >= 0
        super(AppPermission, self).__init__({
            'category': category,
            'id': id,
            'level': level
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['category'], dict_obj['id'], dict_obj['level'])


class AppRole(dict):
    """ Structure of an app role.

    Fields:
        name (str): Role name, refer to 'AppRoleName' enum for valid values.
        permissions (list): List of AppPermission/dict.
    """
    def __init__(self, name, permissions):
        assert name and permissions and isinstance(permissions, list)
        super(AppRole, self).__init__({
            'name': name,
            'permissions': [x if isinstance(x, AppPermission) else AppPermission.from_dict(x) for x in permissions]
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['permissions'])


class Group(dict):
    """ Structure of a group.

    Fields:
        name (str): The group id/name of.
        displayName (str): The group display name.
        id (int): The auto id of the group.
        createTime (long): The group create time.
    """
    def __init__(self, name, display_name, id=None, create_time=None):
        assert name and display_name
        super(Group, self).__init__({
            'name': name,
            'displayName': display_name,
            'id': id,
            'createTime': create_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj.get('id'), dict_obj.get('createTime'))


class UserRole(dict):
    """ Structure of a user role.

    Fields:
        roleName (str): The role of the user in an application. Refer to 'AppRoleName' enum for valid values.
        appId (str): The application id. Refer to 'AppId' enum for valid values.
    """
    def __init__(self, role, app_id):
        assert role and app_id
        super(UserRole, self).__init__({
            'roleName': role,
            'appId': app_id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['roleName'], dict_obj['appId'])


class User(dict):
    """ Structure of a user.

    Fields:
        name (str): The user login name (upper, lower, number, and underscore only).
        displayName (str): The user display name (human-readable).
        password (str): The password of the user.
        department (str): The department of the user (human-readable string).
        mail (str): The e-mail address of the user (human-readable string).
        roles (list): A list of UserRole instances representing the user roles in each application.

        id (int): The auto id of the user.
        createTime (long): The create time in Epoch (milliseconds).
        lastUpdateTime (long): The last update time in Epoch (milliseconds).
    """
    def __init__(self, name, display_name, password=None, department=None, mail=None, roles=None,
                 id=None, create_time=None, last_update_time=None):
        assert name and display_name
        if roles is None:
            roles = []
        assert isinstance(roles, list)
        if department is None:
            department = ''
        if mail is None:
            mail = ''
        super(User, self).__init__({
            'name': name,
            'displayName': display_name,
            'password': password,
            'department': department,
            'mail': mail,
            'roles': [x if isinstance(x, UserRole) else UserRole.from_dict(x) for x in roles],
            'id': id,
            'createTime': create_time,
            'lastUpdateTime': last_update_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj.get('password'), dict_obj.get('department'),
                   dict_obj.get('mail'), dict_obj.get('roles'),
                   dict_obj.get('id'), dict_obj.get('createTime'), dict_obj.get('lastUpdateTime'))


class EventCollector(dict):
    """ Structure of Event Collector data source content.

    Fields:
        hostName (str): The host/domain name of the data source. Set wildcard (*) to accept data from all domains.
    """
    def __init__(self, hostname):
        assert hostname
        super(EventCollector, self).__init__({
            'hostName': hostname
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['hostName'])


class DataSource(dict):
    """ Structure of a data source.

    Fields:
        name (str): The data source id/name.
        displayName (str): The data source display name.
        appIds (list[str]): A list of authorized applications, refer to 'AppId' enum for valid values.
        contentType (str): The data source content type, refer to 'DataSourceContentType' enum for valid values
        type (str): The data source type (how to get data), refer to 'DataSourceType' enum for valid values

        eventCollector (obj): An EventCollector object if type is 'EVENT_COLLECTOR'
        importer (obj): An Importer object if type is 'IMPORTER' (NOT SUPPORTED YET)
        fetch (obj): An Fetch object if type is 'FETCH' (NOT SUPPORTED YET)
        upload (obj): An Upload object if type is 'UPLOAD' (NOT SUPPORTED YET)

        id (int): Data source auto id
        groupId (int): Group auto id
    """
    # TODO: importer, fetch, upload not supported yet
    def __init__(self, name, display_name, app_ids, content_type, id=None, group_id=None):
        super(DataSource, self).__init__({
            'name': name,
            'displayName': display_name,
            'appIds': app_ids,
            'contentType': content_type,
            'id': id,
            'groupId': group_id
        })

    def init_event_collector(self, hostname):
        """ Initialize the data source as Event Collector type.

        Arguments:
            hostname (str): The host/domain name of the data source. Set wildcard (*) to accept data from all domains.
        """
        event_collector = EventCollector(hostname)
        self.update({
            'type': DataSourceType.EVENT_COLLECTOR,
            'eventCollector': event_collector
        })

    def init_event_collector_from_dict(self, dict_obj):
        event_collector = EventCollector.from_dict(dict_obj)
        self.update({
            'type': DataSourceType.EVENT_COLLECTOR,
            'eventCollector': event_collector
        })

    @classmethod
    def from_dict(cls, dict_obj):
        new_obj = cls(dict_obj['name'], dict_obj['displayName'], dict_obj['appIds'], dict_obj['contentType'],
                      dict_obj.get('id'), dict_obj.get('groupId'))

        ds_type = dict_obj.get('type')
        if ds_type == DataSourceType.EVENT_COLLECTOR:
            new_obj.init_event_collector_from_dict(dict_obj['eventCollector'])

        return new_obj


class ExporterExtraSchema(dict):
    """ Structure for Extra Schema in Exporter Setting.

    Fields:
        name (str): Name/field of the schema.
        type (str): Type of the schema, e.g. string, int.
        link (str): The tuple key to explode. Set it to '' for first level event collector parameters.
    """

    def __init__(self, name, type, link):
        assert name and type and link is not None
        super(ExporterExtraSchema, self).__init__({
            'name': name,
            'type': type,
            'link': link
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['type'], dict_obj['link'])


class ExporterSetting(dict):
    """ Structure for Exporter setting.

    Fields:
        enabled (bool): The exporter is enabled or not.
        initalConvetTime (long): The initial conversion time in Epoch (milliseconds). Default: now.
        tupleKey (str): The tuple key. Default: ptuple.
        extraSchema (object): The extra schema.
        parsingFormat (str): The input data format. Default: NginxPlusLogParserDriver.
        baseSchema (obj); The base/default schema.
        updateTime (long): The latest data export time in Epoch (milliseconds).

    NOTE:
        initalConvetTime is the legacy typo of "initialConversionTime".
    """
    def __init__(self, enabled, initial_convert_time=None,
                 tuple_key=None, extra_schema=None,
                 parsing_format=None, base_schema=None, update_time=None):
        assert enabled is not None
        if initial_convert_time is None:
            initial_convert_time = long(time.mktime(time.localtime())*1000)
        if not tuple_key:
            tuple_key = 'ptuple'
        if not extra_schema:
            extra_schema = []
        if not parsing_format:
            parsing_format = 'NginxPlusLogParserDriver'
        if not base_schema:
            base_schema = []

        extra_schema_list = [x if isinstance(x, ExporterExtraSchema) else ExporterExtraSchema.from_dict(x) for x in extra_schema]
        extra_schema_str = json.dumps(extra_schema_list)

        base_schema_list = [x if isinstance(x, ExporterExtraSchema) else ExporterExtraSchema.from_dict(x) for x in base_schema]
        base_schema_str = json.dumps(base_schema_list)

        super(ExporterSetting, self).__init__({
            'enabled': enabled,
            'initalConvetTime': long(initial_convert_time),
            'tupleKey': tuple_key,
            'extraSchema': extra_schema_str,
            'parsingFormat': parsing_format,
            'baseSchema': base_schema_str,
            'updateTime': update_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        extra_schema = None
        if 'extraSchema' in dict_obj:
            extra_schema = json.loads(dict_obj['extraSchema'])
        base_schema = None
        if 'baseSchema' in dict_obj:
            base_schema = json.loads(dict_obj['baseSchema'])
        return cls(dict_obj['enabled'], dict_obj.get('initalConvetTime'),
                   dict_obj.get('tupleKey'), extra_schema,
                   dict_obj.get('parsingFormat'), base_schema, dict_obj.get('updateTime'))


class DataSourceSchema(dict):
    """ Structure for the schema of different data source content type.
    
    Fields:
        name (str): The data field name.
        type (str): The data field type, e.g. string, int
        required (bool): The data field is mandatory or not.

        (For DataSourceContentType.BEHAVIOR)
        link (str): If the data field in "ptuple" or not. "ptuple" is supported yet.

        (For DataSourceContentType.ITEM_INFO, DataSourceContentType.USER_PROFILE, and DataSourceContentType.USER_RANK)
        key (bool): THe data field is key field or not.
    """

    def __init__(self, name, type, required, link=None, key=None):
        assert name and type and required
        super(DataSourceSchema, self).__init__({
            'name': name,
            'type': type,
            'required': True if bool(required) else False,

            # behavior
            'link': link,
            # userprofile, item, userrank
            'key': key
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['type'], dict_obj['required'], dict_obj.get('link'), dict_obj.get('key'))


class SystemInfo(dict):
    """ Structure for system information

    Fields:
        version (str): EMC application version
        build (str): EMC application build number
        branch (str): EMC application build branch
    """

    def __init__(self, version, build, branch):
        assert version and build and branch
        super(SystemInfo, self).__init__({
            'version': version,
            'build': build,
            'branch': branch
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['version'], dict_obj['build'], dict_obj['branch'])


class AuditLogEvent(dict):
    """ Structure for an audit log event.

    Fields:
        eventTime (long): The event time in Epoch (milliseconds).
        userId (int): The user id of the event.
        groupUser (str): A human-readable string of the group/user of the event.
        msg (str): Detail information of the event. Basically in "[Method]xxx;[Arguments]yyy;" format
    """

    def __init__(self, event_time, user_id, group_user, msg):
        assert event_time and user_id and group_user and msg
        super(AuditLogEvent, self).__init__({
            'eventTime': event_time,
            'userId': user_id,
            'groupUser': group_user,
            'msg': msg
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['eventTime'], dict_obj['userId'], dict_obj['groupUser'], dict_obj['msg'])


class AuditLog(dict):
    """ Structure for audit log.
    
    Fields:
        startTime (long): The start time in the audit log query in Epoch (milliseconds).
        endTime (long): The end time in the audit log query in Epoch (milliseconds).
        events (list): A list of "AuditLogEvent" instances.
    """

    def __init__(self, start_time, end_time, events):
        assert start_time and end_time and events is not None and isinstance(events, list)
        super(AuditLog, self).__init__({
            'startTime': start_time,
            'endTime': end_time,
            'events': [AuditLogEvent.from_dict(x) for x in events]
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['startTime'], dict_obj['endTime'], dict_obj['events'])


class EMC2(BaseApp):
    """ Encapsulate Etu Management Center (v2) API. """

    __APP_NAME = 'EMC2'
    __HOST = 'emc.online.etunexus.com'
    __API_BASE = '/commsrv/v1'
    __SHIRO_CAS_BASE = '/shiro-cas'

    def __init__(self, cas, host=None, api_base=None, shiro_cas_base=None):
        """ Constructor of EMC2 instance. """

        api_base = api_base if api_base else self.__API_BASE
        super(EMC2, self).__init__(cas, EMC2.__APP_NAME,
                                   api_host=host if host else self.__HOST,
                                   api_base=api_base if api_base else self.__API_BASE,
                                   shiro_cas_base=shiro_cas_base if shiro_cas_base else self.__SHIRO_CAS_BASE)

    # Group #
    def get_groups(self):
        """ Get group list.

        Arguments:
        Return:
            A list of Group instances.
        """
        res = self.request_get('/group')
        return [Group.from_dict(x) for x in res]

    def add_group(self, group):
        """ Add a new group.

        Arguments:
            group (obj): The Group instance to add.
        Return:
            A Group instance as the added one (with fields filled by server, e.g. createTime).
        """
        assert group and isinstance(group, Group)
        res = self.request_post('/group', group)
        return Group.from_dict(res)

    def update_group(self, group):
        """ Update a group.

        Arguments:
            group (obj): The Group instance to update, with valid "id".
        Return:
            A Group instance as the updated one.
        """
        assert group and isinstance(group, Group)
        group_id = group['id']
        res = self.request_post('/group/{0}'.format(group_id), group)
        return Group.from_dict(res)

    def del_group(self, group):
        """ Delete a group.

        Arguments:
            group (obj or int): The Group instance or group id to delete.
        Return:
            The group id deleted. It must be the same as the one in argument.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) else int(group)
        assert group_id
        res = self.request_del('/group/{0}'.format(group_id))
        assert res == group_id
        return res

    # User #
    def me(self):
        """ Get "me" user instance

        Arguments:
        Return:
            A User instance as "me", the login user or simulated user
        """
        res = self.request_get('/user/me')
        return User.from_dict(res)

    def get_users(self, group):
        """ Get user list in a group.

        Arguments:
            group (obj or int): The Group instance or group id to get user list.
        Return:
            A list of User instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) else int(group)
        res = self.request_get('/group/{0}/user'.format(group_id))
        return [User.from_dict(x) for x in res]

    def add_user(self, group, user):
        """ Add a user into a group.

        Arguments:
            group (obj or int): The Group instance or group id to add the user into.
            user (obj): The User instance to add.
        Return:
            An User instance as the added one (with fields filled by server, e.g. createTime).
        """
        assert group and user and isinstance(user, User)
        group_id = group['id'] if isinstance(group, Group) else int(group)
        res = self.request_post('/group/{0}/user'.format(group_id), user)
        return User.from_dict(res)

    def update_user(self, user):
        """ Update an user.

        Do not use this method to change user password. Use "change_user_password()" instead.

        Arguments:
            user (obj): The User instance to update, with a valid "id".
        Return:
            An User instance as the updated one.
        """
        assert user and isinstance(user, User)
        user_id = user['id']
        res = self.request_post('/user/{0}'.format(user_id), user)
        return User.from_dict(res)

    def del_user(self, user):
        """ Delete an user.

        Arguments:
            user (obj or int): The User instance or user id to delete.
        Return:
            The user id deleted. It must be the same as the one in argument.
        """
        assert user
        user_id = user['id'] if isinstance(user, User) else int(user)
        res = self.request_del('/user/{0}'.format(user_id))
        assert res == user_id
        return res

    def change_user_password(self, user, password):
        """ Change user password.

        For normal user, the "user" argument should be "me".
        For operator, the "user" can be another user in the same group.
        For admin, The "user" can be anyone in the system.

        Arguments:
            user (obj or int): The User instance or user id to change password.
            password (str): The new password to set.
        Return:
            The user id with password changed. It must be the same as the one in argument.
        """
        assert user and password
        user_id = user['id'] if isinstance(user, User) else int(user)
        params = {
            'password': password
        }
        res = self.request_post('/user/{0}/password'.format(user_id), params)
        assert res == user_id
        return res

    # Apps #
    def get_apps(self):
        """ Get application list supported in the system.

        Arguments:
        Return:
            A list of AppInfo instances as the applications installed in system.
        """
        res = self.request_get('/app')
        return [AppInfo.from_dict(x) for x in res]

    def get_app_permission(self, app_id):
        """ Get application permission settings.

        Arguments:
            app_id (str): An application id, refer to "AppId" enum for the valid values.
        Return:
            A list of AppPermission instances as the setting of the given application.
        """
        assert app_id
        res = self.request_get('/app/{0}/permission'.format(app_id))
        assert (res['appId'] == app_id)
        return [AppPermission.from_dict(x) for x in res['permissions']]

    def get_app_roles(self, app_id):
        """ Get application role and permission settings.

        Arguments:
            app_id (str): An application id, refer to "AppId" enum for the valid values.
        Return:
            A list of AppRole instances as the role setting of the given application.
        """
        assert app_id
        res = self.request_get('/app/{0}/role'.format(app_id))
        return [AppRole.from_dict(x) for x in res]

    def update_app_role(self, app_id, app_role):
        """ Update a role and permission setting in an application.

        Arguments:
            app_id (str): An application id, refer to "AppId" enum for the valid values.
            app_role (obj): The AppRole instance to update.
        Return:
            An AppRole as the as the updated one.
        """
        assert app_id and app_role and isinstance(app_role, AppRole)
        res = self.request_post('/app/{0}/role'.format(app_id), app_role)
        return AppRole.from_dict(res)

    # Data source #
    def get_data_sources(self, group):
        """ Get data source list in a group.

        Arguments:
            group (obj or int): The Group instance or group id to get.
        Return:
            A list of DataSource instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) else int(group)
        res = self.request_get('/group/{0}/data-source'.format(group_id))
        return [DataSource.from_dict(x) for x in res]

    def add_data_source(self, group, data_source):
        """ Add a data source to a group.

        To add a new data source, first create and initialize a DataSource instance before calling. During initializing
        data source, not only "new" an instance but need to call the proper "init_xxx()" method as requested.

        Arguments:
            group (obj or int): The Group instance or group id to get.
            data_source (object): The DataSource instance to add.
        Return:
            A DataSource instance as the added one.
        """
        assert data_source and isinstance(data_source, DataSource)
        group_id = group['id'] if isinstance(group, Group) else int(group)
        res = self.request_post('/group/{0}/data-source'.format(group_id), data_source)
        return DataSource.from_dict(res)

    def update_data_source(self, data_source):
        """ Update a data source.

        Arguments:
            data_source (obj): The DataSource instance to update, with valid "id".
        Return:
            A DataSource instance as the updated one.
        """
        assert data_source and isinstance(data_source, DataSource)
        source_id = data_source['id']
        assert source_id
        res = self.request_post('/data-source/{0}'.format(source_id), data_source)
        return DataSource.from_dict(res)

    def del_data_source(self, data_source):
        """ Delete a data source.

        Arguments:
            data_source (obj or int): The DataSource instance or data source id to delete.
        Return:
            The data source id deleted. It must be the same as the one in argument.
        """
        assert data_source
        source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        assert source_id
        res = self.request_del('/data-source/{0}'.format(source_id))
        assert res == source_id
        return res

    # Exporter setting #
    def get_exporter_setting(self, data_source):
        """ Get Exporter setting of an event collector type data source.

        As the input argument accepts a generic data source id, this function does not check if the data source is
        really "event collector" type. The caller SHOULD ensure the data source is eligible for configuring exporter.

        Arguments:
            data_source (obj or int): The DataSource instance or data source id to get exporter setting.
        Return:
            An ExporterSetting instance.
        """
        assert data_source
        source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        res = self.request_get('/data-source/{0}/exporter'.format(source_id))
        return ExporterSetting.from_dict(res)

    def update_exporter_setting(self, data_source, exporter_setting):
        """ Update Exporter setting of an event collector type data source.

        As the input argument accepts a generic data source id, this function does not check if the data source is
        really "event collector" type. The caller SHOULD ensure the data source is eligible for configuring exporter.

        Arguments:
            data_source (obj or int): The DataSource instance or data source id to get exporter setting.
            exporter_setting (obj): The ExporterSetting instance to apply to the data source.
        Return:
            An ExporterSetting instance as the applied one.
        """
        assert data_source and  exporter_setting and isinstance(exporter_setting, ExporterSetting)
        source_id = data_source['id'] if isinstance(data_source, DataSource) else int(data_source)
        assert source_id
        res = self.request_post('/data-source/{0}/exporter'.format(source_id), exporter_setting)
        return ExporterSetting.from_dict(res)

    # Data source information #
    def get_data_source_schema(self, content_type):
        """ Get data source schema.
        
        Arguments:
            content_type (str): The data source content type, refer to DataSourceContentType for valid values.
        Return:
            A list of DataSourceSchema instances.
        """
        assert content_type
        res = self.request_get('/data-source/schema/{0}'.format(content_type))
        return [DataSourceSchema.from_dict(x) for x in res]

    # System #
    def get_system_info(self):
        """ Get EMC application information (system information).

        Currently, only version information meaningful.

        Arguments:
        Return:
            A SystemInfo instance.
        """
        res = self.request_get('/system/version')
        return SystemInfo.from_dict(res)

    def get_audit_logs(self, start_time, end_time):
        assert start_time and end_time
        res = self.request_get('/audit?startTime={0}&endTime={1}'.format(start_time, end_time))
        return AuditLog.from_dict(res)


# -*- coding: utf-8 -*-

import logging


def get_logger(level=logging.INFO):
    logger = logging.getLogger('etu.nexus')
    logger.setLevel(level)
    if not len(logger.handlers):
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(module)s:%(lineno)d - %(message)s'))
        logger.addHandler(ch)

    return logger



# -*- coding: utf-8 -*-

from baseapp import BaseApp
from emc import Group, DataSource
from enum import *


class ERGroup(dict):
    """ Structure of a ER specific group.

    NOTICE: No createTime field as in emc.Group.

    Fields:
        name (str): The group id/name of.
        displayName (str): The group display name.
        id (int): The auto id of the group.
        createTime (long): The group create time.
    """
    def __init__(self, name, display_name, id=None):
        assert name and display_name
        super(ERGroup, self).__init__({
            'name': name,
            'displayName': display_name,
            'id': id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj.get('id'))


class ERDataSource(dict):
    """ Structure of a ER specific/authorized data source.

    NOTICE: While all data source management should be applied on EMC. The structure here includes only basic
    information as not as rich as emc.DataSource.

    Fields:
        name (str): The data source id/name.
        displayName (str): The data source display name.
        contentType (str): The data source content type, refer to 'DataSourceContentType' enum for valid values
        type (str): The data source type (how to get data), refer to 'DataSourceType' enum for valid values

        id (int): Data source auto id
        groupId (int): Group auto id
    """
    def __init__(self, name, display_name, content_type, type, id=None, group_id=None):
        super(ERDataSource, self).__init__({
            'name': name,
            'displayName': display_name,
            'contentType': content_type,
            'type': type,

            'id': id,
            'groupId': group_id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj['contentType'], dict_obj['type'],
                   dict_obj.get('id'), dict_obj.get('groupId'))


class AlgInstance(dict):
    """ Structure of an algorithm instance in the recommendation logic.

    AlgInstance is an ABC for all different algorithm structure. The inherited algorithm implementation should be named
    as "Alg_[algId]", e.g. Alg_USER_BASED_CF or Alg_ALS.

    Fields:
        algId (str): The algorithm id, every derived algorithm should have its own id, refer to "LogicAlgorithmId" enum
        for samples but not limited.
        weight (int): The weight of the algorithm in the linear combination to the recommendation logic result.
        setting (obj): The setting depend on the concrete algorithm implementation.

        id (int): The auto id
        executeTime (long): Last execution/processing time in Epoch (milliseconds).
        executeState (str): Last execution status, refer to "LogicExecuteState" enum for valid values.
        successTime (long): Last success execution time in Epoch (milliseconds).
    """
    def __init__(self, alg_id, weight, setting):
        assert alg_id and weight and setting and isinstance(setting, dict)
        super(AlgInstance, self).__init__({
            'algId': alg_id,
            'weight': weight,
            'setting': setting
        })

    def __init_general(self, dict_obj):
        self.update({
            'id': dict_obj.get('id'),
            'executeTime': dict_obj.get('executeTime'),
            'executeState': dict_obj.get('executeState'),
            'successTime': dict_obj.get('successTime')
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        obj = cls._create_alg_instance(dict_obj['algId'], dict_obj)
        if obj is not None:
            obj.__init_general(dict_obj)
        return obj

    @classmethod
    def _create_alg_instance(cls, alg_id, dict_obj):
        alg_class_name = 'Alg_{0}'.format(alg_id)
        try:
            klass = globals()[alg_class_name]
        except KeyError, ke:
            klass = None
        return klass.from_dict(dict_obj) if klass else Alg_UNKNOWN.from_dict(dict_obj)


class Alg_UNKNOWN(AlgInstance):
    """ Structure of unknown/not-supported algorithm.

    Fields:
        (As 'setting' in AlgInstance)
    """
    def __init__(self, weight, setting):
        assert weight and setting and isinstance(setting, dict)
        super(Alg_UNKNOWN, self).__init__(LogicAlgorithmId.UNKNOWN, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting)


class Alg_USER_BASED_CF(AlgInstance):
    """ Structure of user-based CF algorithm.

    Fields:
        (As "setting" in AlgInstance)
        id (str): Fixed "USER_BASED_CF".
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TIMERANGE (int): The data time range to calculate.
        action (list): A list of event actions to calculate, refer to "EventAction" enum for samples but not limited.
        als_iteration (int): ALS iterations.
        als_rank (int): ALS training features.
    """
    def __init__(self, weight, data_source, time_range, actions, als_iteration=10, als_rank=10):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert actions and isinstance(actions, list)
        setting = {
            'id': LogicAlgorithmId.USER_BASED_CF,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TIMERANGE': time_range,
            'action': actions,
            'als_iteration': als_iteration,
            'als_rank': als_rank
        }
        super(Alg_USER_BASED_CF, self).__init__(LogicAlgorithmId.USER_BASED_CF, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], setting['TIMERANGE'], setting['action'],
                   setting.get('als_iteration'), setting.get('als_rank'))


class Alg_ITEM_BASED_CF(AlgInstance):
    """ Structure of item-based CF algorithm.

    Fields:
        (As 'setting' in AlgInstance)
        id (str): Fixed "ITEM_BASED_CF".
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TIMERANGE (int): The data time range to calculate.
        action (list): A list of event actions to calculate, refer to "EventAction" enum for samples but not limited.
        als_iteration (int): ALS iterations.
        als_rank (int): ALS training features.
    """
    def __init__(self, weight, data_source, time_range, actions, als_iteration=10, als_rank=10):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert actions and isinstance(actions, list)
        setting = {
            'id': LogicAlgorithmId.ITEM_BASED_CF,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TIMERANGE': time_range,
            'action': actions,
            'als_iteration': als_iteration,
            'als_rank': als_rank
        }
        super(Alg_ITEM_BASED_CF, self).__init__(LogicAlgorithmId.ITEM_BASED_CF, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], setting['TIMERANGE'], setting['action'],
                   setting.get('als_iteration'), setting.get('als_rank'))


class Alg_RANKING(AlgInstance):
    """ Structure of ranking algorithm.

    Fields:
        (As 'setting' in AlgInstance)
        id (str): Fixed "RANKING".
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TIMERANGE (int): The data time range to calculate.
        actionList (list): A list of event actions to calculate, refer to "EventAction" enum for samples but not
        limited.
        addNonCategoryRec (bool): Generate category-independent recommendation list or not.
        delimiter (str): The delimiter in the category string for multiple levels categories.
        flatAct (bool): Make the category as hierarchy (false, default) or flat (true).
    """
    def __init__(self, weight, data_source, time_range, actions, gen_non_category_rec=True, delimiter=None,
                 flat_cat=False):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert actions and isinstance(actions, list)
        if delimiter is None:
            delimiter = ','

        setting = {
            'id': LogicAlgorithmId.RANKING,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TIMERANGE': time_range,
            'actionList': actions,
            'addNonCategoryRec': str(bool(gen_non_category_rec)).lower(),
            'delimiter': delimiter,
            'flatAct': flat_cat
        }
        super(Alg_RANKING, self).__init__(LogicAlgorithmId.RANKING, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], setting['TIMERANGE'], setting['actionList'],
                   setting.get('addNonCategoryRec'), setting.get('delimiter'), setting.get('flatAct'))


class Alg_RANKING_ITEMINFO(AlgInstance):
    """ Structure for ranking with item info algorithm.

    Fields:
        (As 'setting' in AlgInstance)
        id: Fixed "RANKING_ITEMINFO".
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TIMERANGE (int): The data time range to calculate.
        actionList (list): A list of event actions to calculate, refer to "EventAction" enum for samples but not
        limited.
        datasource_Iteminfo (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and
        "name").
        addNonCategoryRec (bool): Generate category-independent recommendation list or not.
        delimiter (str): The delimiter in the category string for multiple levels categories.
        flatAct (bool): Make the category as hierarchy (false, default) or flat (true).
    """
    def __init__(self, weight, data_source, time_range, actions, item_data_source,
                 gen_non_category_rec=True, delimiter=None, flat_cat=False):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert actions and isinstance(actions, list)
        assert item_data_source and isinstance(item_data_source, dict)
        if delimiter is None:
            delimiter = ','

        setting = {
            'id': LogicAlgorithmId.RANKING,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TIMERANGE': time_range,
            'actionList': actions,
            'datasource_Iteminfo': {'id': item_data_source['id'], 'name': item_data_source['name']},
            'addNonCategoryRec': str(bool(gen_non_category_rec)).lower(),
            'delimiter': delimiter,
            'flatAct': flat_cat
        }
        super(Alg_RANKING_ITEMINFO, self).__init__(LogicAlgorithmId.RANKING_ITEMINFO, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], setting['TIMERANGE'], setting['actionList'],
                   setting['datasource_Iteminfo'],
                   setting.get('addNonCategoryRec'), setting.get('delimiter'), setting.get('flatAct'))


class Alg_SEARCH2CLICK(AlgInstance):
    """ Structure for search to click algorithm.

    Fields:
        (As 'setting' in AlgInstance)
        id: Fixed "SEARCH2CLICK"
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TIMERANGE (int): The data time range to calculate.
        TIME_DECAYED (int): The time decay factor.
        ACTION (str): The event action to calculate after the search event.
    """
    def __init__(self, weight, data_source, time_range, time_decay_factor, action):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert time_decay_factor
        assert action
        setting = {
            'id': LogicAlgorithmId.SEARCH2CLICK,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TIMERANGE': time_range,
            'TIME_DECAYED': time_decay_factor,
            'ACTION': action
        }
        super(Alg_SEARCH2CLICK, self).__init__(LogicAlgorithmId.SEARCH2CLICK, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], setting['TIMERANGE'], setting['TIME_DECAYED'],
                   setting['ACTION'])


class Alg_Info_Integrity(AlgInstance):
    """ Structure for item info integrity recommendation.
    
    Fields:
        (As 'setting' in AlgInstance)
        id: Fixed "Info_Integrity"
        dataSource (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        attritubeList (list): The attributes used to process integrity recommendation.

    Notice: "attritubeList" is the typo of "attributeList".
    """
    def __init__(self, weight, data_source, attributes):
        assert data_source and isinstance(data_source, dict)
        assert attributes and isinstance(attributes, list)
        setting = {
            'id': LogicAlgorithmId.INFO_INTEGRITY,
            'dataSource': {'id': data_source['id'], 'name': data_source['name']},
            'attritubeList': attributes
        }
        super(Alg_Info_Integrity, self).__init__(LogicAlgorithmId.INFO_INTEGRITY, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['dataSource'], setting['attritubeList'])


class Alg_ALS(AlgInstance):
    """ Structure for ALS recommendation.

    Fields:
        (As 'setting' in AlgInstance)
        id: Fixed "ALS"
        dataSource (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        timeRange (int): The data time range to calculate.
        LAMBDA (float): The lambda value for ALS algorithm.
    """
    def __init__(self, weight, data_source, time_range, lambda_val=0.1):
        assert data_source and isinstance(data_source, dict)
        assert time_range
        assert lambda_val
        setting = {
            'id': LogicAlgorithmId.ALS,
            'dataSource': {'id': data_source['id'], 'name': data_source['name']},
            'timeRange': time_range,
            'LAMBDA': lambda_val
        }
        super(Alg_ALS, self).__init__(LogicAlgorithmId.ALS, weight, setting)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['dataSource'], setting['timeRange'], setting['LAMBDA'])


class Alg_LDA(AlgInstance):
    """ Structure for LDA context-aware recommendation.

    Fields:
        (As 'setting' in AlgInstance)
        id: Fixed "LDA"
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        applyTrainingId (obj or int): An AlgTraining instance, a dict instance with "id", or a number.
    """
    def __init__(self, weight, data_source, apply_training_id=None):
        assert data_source and isinstance(data_source, dict)
        setting = {
            'id': LogicAlgorithmId.LDA,
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']}
        }
        super(Alg_LDA, self).__init__(LogicAlgorithmId.LDA, weight, setting)
        # It's an work-around for the bad API design that put "applyTrainingId" in general algorithm instance level
        # instead of setting for specific algorithm.
        if apply_training_id is not None:
            apply_training_id = apply_training_id['id'] if isinstance(apply_training_id, dict) \
                else int(apply_training_id)
        self.update({
            'applyTrainingId': apply_training_id
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        return cls(dict_obj['weight'], setting['DATASOURCE'], dict_obj['applyTrainingId'])


class AlgTraining(dict):
    """ Structure for a algorithm training setting and result.
    
    Fields:
        algId (str): The algorithm id, every derived algorithm should have its own id, refer to "LogicAlgorithmId" enum
        for samples but not limited.
        setting (obj): The setting depend on the concrete algorithm implementation.
        state (str): The training execution state, refer to "AlgTrainingState" enum for valid values.
        result (obj): The result depend on the concrete algorithm implementation.
        
        id (int): The auto id.
        createTime (long): Create time in Epoch (milliseconds).
        updateTime (long): Update time in Epoch (milliseconds).
    """

    def __init__(self, alg_id, setting):
        assert alg_id and setting and isinstance(setting, dict)
        super(AlgTraining, self).__init__({
            'algId': alg_id,
            'setting': setting
        })

    def __init_result__(self, result):
        self.update({'result': result})

    def __init_general__(self, dict_obj):
        self.update({
            'id': dict_obj['id'],
            'state': dict_obj['state'],
            'createTime': dict_obj['createTime'],
            'updateTime': dict_obj['updateTime']
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        obj = cls._create_algtraing_instance(dict_obj['algId'], dict_obj)
        if obj is not None:
            obj.__init_general__(dict_obj)
        return obj

    @classmethod
    def _create_algtraing_instance(cls, alg_id, dict_obj):
        alg_class_name = 'AlgTraining_{0}'.format(alg_id)
        try:
            klass = globals()[alg_class_name]
        except KeyError, ke:
            klass = None
        return klass.from_dict(dict_obj) if klass else AlgTraining_UNKNOWN.from_dict(dict_obj)


class AlgTraining_UNKNOWN(AlgTraining):
    """ Structure of unknown/not-supported algorithm training result.

    Fields:
        (As 'setting' in AlgTraining)
        (as 'result' in AlgTraining)
    """
    def __init__(self, setting):
        assert setting and isinstance(setting, dict)
        super(AlgTraining_UNKNOWN, self).__init__(LogicAlgorithmId.UNKNOWN, setting)

    def __init_result__(self, result):
        super(AlgTraining_UNKNOWN, self).__init_result__(result)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        obj = cls(setting)
        result = dict_obj.get('result')
        if result is not None:
            obj.__init_result(result)
        return obj


class AlgTraining_LDA(AlgTraining):
    """ Structure for LDA context-aware recommendation model training.

    Fields:
        (As 'setting' in AlgTraining)
        DATASOURCE (obj): An emc.DataSource instance, ERDataSource instance, or a dict instance with "id" and "name").
        TOPICS (int): The number of topics.
        
        (As 'result' in AlgTraining)
        dataSource (str): Only the data source name.
        ldaModalPath (str): The path on HDFS keeps the training model.
    """
    def __init__(self, data_source, topics):
        assert data_source and isinstance(data_source, dict)
        assert topics and isinstance(topics, int)
        setting = {
            'DATASOURCE': {'id': data_source['id'], 'name': data_source['name']},
            'TOPICS': topics
        }
        super(AlgTraining_LDA, self).__init__(LogicAlgorithmId.LDA, setting)

    def __init_result__(self, result):
        assert result is not None and isinstance(result, dict)
        final_result = {}
        if len(result) > 0:
            final_result = {
                'dataSource': result['dataSource'],
                'ldaModalPath': result['ldaModalPath']
            }
        super(AlgTraining_LDA, self).__init_result__(final_result)

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj and dict_obj['setting']
        setting = dict_obj['setting']
        obj = cls(setting['DATASOURCE'], setting['TOPICS'])
        result = dict_obj.get('result')
        if result is not None:
            obj.__init_result__(result)
        return obj


class UserFilter(dict):
    """ Structure of a user filter setting.

    Fields:
        name (str): The user filter name/id.
        displayName (str): The user filter display name.

        id (int): The auto id.
        createTime (long): Create time in Epoch (milliseconds).
        updateTime (long): Update time in Epoch (milliseconds).
    """
    def __init__(self, name, display_name, id=None, create_time=None, update_time=None):
        assert name and display_name
        super(UserFilter, self).__init__({
            'name': name,
            'displayName': display_name,
            'id': id,
            'createTime': create_time,
            'updateTime': update_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        assert dict_obj
        return cls(dict_obj['name'], dict_obj['displayName'],
                   dict_obj.get('id'), dict_obj.get('createTime'), dict_obj.get('updateTime'))


class Logic(dict):
    """ Structure of a recommendation logic.

    Fields:
        name (str): The logic name/id.
        displayName (str): The logic display name.
        active (bool): The logic is active or not.
        numberOfRec (int): Max count of recommendation results.
        algType (str): The logic algorithm type, refer to "LogicAlgType" enum for valid values.
        algInstances (list): Algorithm instances in the logic.
        userFilter (object): User filter setting.
        useLocation (bool): Use location or not.
        delegateLogicName (str): The logic(s) for recommendation complementary.
        filteringLogicIds (list): A list of "Logic" instances or logic IDs.
        enableUpdating (bool): Enable core engine process or not.

        ## Item Filtering (only for altType==ITEM_BASE)
        enableLastViewedItem (bool): Enable last viewed item recommendation or not.
        itemFilterSrc (str): The item info data source used for filtering.
        enableSameCategory (bool): Enable item in same category recommendation or not.
        avlItemFilterMode (str): The item avl filter mode, refer to "LogicAvlItemFilterMode" enum for valid values.

        id (int): The auto id.
        createTime (long): Create time in Epoch (milliseconds).
        updateTime (long): Update time in Epoch (milliseconds).
        executeTime (long): Last execution/processing time in Epoch (milliseconds).
        executeState (str): Last execution status, refer to "LogicExecuteState" enum for valid values.
        successTime (long): Last success execution time in Epoch (milliseconds).
    """
    def __init__(self, name, display_name, active, rec_count,
                 alg_type, alg_instances,
                 user_filter=None,
                 use_location=False, enable_last_viewed_item=False,
                 complementary_logics=None, filtering_logics=None, enable_updating=True,
                 item_filter_src=None, enable_same_category=False, avl_item_filter_mode=LogicAvlItemFilterMode.DISABLED,
                 id=None, create_time=None, update_time=None, execute_time=None, execute_state=None, success_time=None):
        assert name and display_name and rec_count
        assert active is not None and isinstance(active, bool)
        assert alg_type and alg_instances and isinstance(alg_instances, list)
        if user_filter is not None:
            assert isinstance(user_filter, UserFilter) or isinstance(user_filter, dict)
            user_filter = user_filter if isinstance(user_filter, UserFilter) else UserFilter.from_dict(user_filter)
        if filtering_logics is not None:
            assert isinstance(filtering_logics, list)
            if len(filtering_logics) > 0 and isinstance(filtering_logics[0], Logic):
                filtering_logics = [x['id'] for x in filtering_logics]
        else:
            filtering_logics = []
        assert isinstance(use_location, bool)
        assert isinstance(enable_last_viewed_item, bool)
        assert isinstance(enable_same_category, bool)
        super(Logic, self).__init__({
            'name': name,
            'displayName': display_name,
            'active': active,
            'numberOfRec': rec_count,
            'algType': alg_type,
            'algInstances': [x if isinstance(x, AlgInstance) else AlgInstance.from_dict(x) for x in alg_instances],
            'userFilter': user_filter,
            'useLocation': use_location,
            'enableLastViewedItem': enable_last_viewed_item,
            'delegateLogicName': complementary_logics,
            'filteringLogicIds': filtering_logics,
            'enableUpdating': enable_updating,

            'itemFilterSrc': item_filter_src,
            'enableSameCategory': enable_same_category,
            'avlItemFilterMode': avl_item_filter_mode,

            'id': id,
            'createTime': create_time,
            'updateTime': update_time,
            'executeTime': execute_time,
            'executeState': execute_state,
            'successTime': success_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['name'], dict_obj['displayName'], dict_obj['active'], dict_obj['numberOfRec'],
                   dict_obj['algType'], dict_obj['algInstances'],
                   dict_obj.get('userFilter'),
                   dict_obj['useLocation'], dict_obj['enableLastViewedItem'],
                   dict_obj.get('delegateLogicName'), dict_obj.get('filteringLogicIds'), dict_obj['enableUpdating'],
                   dict_obj.get('itemFilterSrc'), dict_obj['enableSameCategory'], dict_obj['avlItemFilterMode'],
                   dict_obj.get('id'), dict_obj.get('createTime'), dict_obj.get('updateTime'),
                   dict_obj.get('executeTime'), dict_obj.get('executeState'), dict_obj.get('successTime'))


class Campaign(dict):
    """ Structure for a recommendation campaign

    Fields:
        name (str): Campaign name/id.
        displayName (str): Campaign display name.
        startTime (long): Campaign start time in Epoch (milliseconds).
        endTime (long): Campaign end time in Epoch (milliseconds).
        logics (list): A list of "Logic" instances.

        id (int): The auto id.
        createTime (long): Campaign create time in Epoch (milliseconds).
        updateTime (long): Campaign update time in Epoch (milliseconds).
    """
    def __init__(self, group_id, name, display_name, start_time, end_time, logics=None, id=None, create_time=None, update_time=None):
        assert group_id and name and display_name
        assert 0 <= start_time and 0 <= end_time and start_time < end_time
        if logics is None:
            logics = []
        super(Campaign, self).__init__({
            'groupId': group_id,
            'name': name,
            'displayName': display_name,
            'startTime': start_time,
            'endTime': end_time,
            'logics': [x if isinstance(x, Logic) else Logic.from_dict(x) for x in logics],
            'id': id,
            'createTime': create_time,
            'updateTime': update_time
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['groupId'], dict_obj['name'], dict_obj['displayName'],
                   dict_obj['startTime'], dict_obj['endTime'],
                   dict_obj['logics]'],
                   dict_obj.get('id'), dict_obj.get('createTime'), dict_obj.get('updateTime'))


class Layout(dict):
    """ Structure for a recommendation layout generator.

    Fields:
        dataSrcName (str): The item info data source name/id.
        loName (str): The layout name/id.
        loTitle (str): The title of the layout.
        loTitleAlign (str): Alignment of the title, refer to "LayoutTitleAlignment" for valid values.
        loItemCnt (int): The total count of items.
        loItemWidth (int): The width of an item (px).
        loItemHeight (int): The height of an item (px).
        loItemMargin (int): The margin between items (px).
        loBGColor (str): Background color in "rgba(r,g,b,alpha)" format.
        loFGColor (str): Foreground color in "rgba(r,g,b,alpha)" format.
        loCol (int): Number of columns.
        loRow (int): Number of rows.
        loFontSize (int): Font size (px).

        loID (int): The auto id
    """

    def __init__(self, data_source, name,
                 title, title_align=LayoutTitleAlignment.CENTER,
                 item_count=10, item_width=200, item_height=200, item_margin=20,
                 bg_color='rgba(255,255,255,0)', fg_color='rgba(0,0,0,1)',
                 columns=5, rows=2, font_size=16):
        assert data_source and name and title
        data_source_name = data_source['name'] \
            if isinstance(data_source, DataSource) and DataSource['contentType'] == DataSourceContentType.ITEM_INFO \
            else str(data_source)
        super(Layout, self).__init__({
            'dataSrcName': data_source_name,
            'loName': name,
            'loTitle': title,
            'loTitleAlign': title_align,
            'loItemCnt': item_count,
            'loItemWidth': item_width,
            'loItemHeight': item_height,
            'loItemMargin': item_margin,
            'loBGColor': bg_color,
            'loFGColor': fg_color,
            'loCol': columns,
            'loRow': rows,
            'loFontSize': font_size
        })

    @classmethod
    def from_dict(cls, dict_obj):
        return cls(dict_obj['dataSrcName'], dict_obj['loName'],
                   dict_obj['loTitle'], dict_obj.get('loTitleAlign'),
                   dict_obj.get('loItemCnt'), dict_obj.get('loItemWidth'), dict_obj.get('loItemHeight'),
                   dict_obj.get('loItemMargin'),
                   dict_obj.get('loBGColor'), dict_obj.get('loFGColor'),
                   dict_obj.get('loCol'), dict_obj.get('loRow'), dict_obj.get('loFontSize'))


class ER3(BaseApp):
    """ Encapsulate Etu Recommender (v3) API """

    __APP_NAME = 'ER3'
    __HOST = 'erhome.online.etunexus.com'
    __API_BASE = '/ER/V3/recsrv/v1'
    __SHIRO_CAS_BASE = '/ER/V3/shiro-cas'

    def __init__(self, cas, host=None, api_base=None, shiro_cas_base=None):
        """ Constructor of ER3 instance. """
        api_base = api_base if api_base else self.__API_BASE
        super(ER3, self).__init__(cas, ER3.__APP_NAME,
                                  api_host=host if host else self.__HOST,
                                  api_base=api_base if api_base else self.__API_BASE,
                                  shiro_cas_base=shiro_cas_base if shiro_cas_base else self.__SHIRO_CAS_BASE)

    # Group #
    def get_groups(self):
        """ Get group (list).

        Depending on the authorization of the user, the group(s) returned would be different. For a normal user, it is
        most likely only the group which the user belongs to.

        Arguments:
        Return:
            A list of ERGroup instances.
        """
        res = self.request_get('/group')
        return [ERGroup.from_dict(x) for x in res]

    # Data source #
    def get_data_sources(self, group):
        """ Get data source list in a group.

        Only the data sources in the group authorized to ER are returned, and it may be just a subset of the list
        returned from the same API in EMC.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to get.
        Return:
            A list of ERDataSource instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_get('/group/{0}/data-source'.format(group_id))
        return [ERDataSource.from_dict(x) for x in res]

    # Logic #
    def get_logics(self, group):
        """ Get the recommendation logics in a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to get.
        Return:
            A list of Logic instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_get('/group/{0}/logic'.format(group_id))
        return [Logic.from_dict(x) for x in res]

    def add_logic(self, group, logic):
        """ Add a new recommendation logic to a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to add logic.
            logic (obj): The Logic instance to add.
        Return:
            A Logic instance as the added one (with fields filled by server, e.g. createTime).
        """
        assert group and logic and isinstance(logic, Logic)
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_post('/group/{0}/logic'.format(group_id), logic)
        return Logic.from_dict(res)

    def update_logic(self, logic):
        """ Update a recommendation logic.

        Arguments:
            logic (obj): The Logic instance to update, with a valid "id".
        Return:
            A Logic instance as the updated one.
        """
        assert logic and isinstance(logic, Logic)
        logic_id = logic['id']
        assert logic_id
        res = self.request_post('/logic/{0}'.format(logic_id), logic)
        return Logic.from_dict(res)

    def del_logic(self, logic):
        """ Delete a recommendation logic.

        Arguments:
            logic (obj or int): The Logic instance or logic id to delete.
        Return:
            The logic id deleted. It must be the same as the one in argument.
        """
        assert logic
        logic_id = logic['id'] if isinstance(logic, Logic) else int(logic)
        assert logic_id
        res = self.request_del('/logic/{0}'.format(logic_id))
        assert res == logic_id
        return res

    # Campaign #
    def get_campaigns(self, group):
        """ Get the campaigns in a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to get.
        Return:
            A list of Campaign instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_get('/group/{0}/campaign'.format(group_id))
        return [Campaign.from_dict(x) for x in res]

    def add_campaign(self, group, campaign):
        """ Add a new campaign to a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to add campaign.
            campaign (obj): The Campaign instance to add.
        Return:
            A Campaign instance as the added one (with fields filled by server, e.g. createTime).
        """
        assert group and campaign and isinstance(campaign, Campaign)
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_post('/group/{0}/campaign'.format(group_id), campaign)
        return Campaign.from_dict(res)

    def update_campaign(self, campaign):
        """ Update a campaign.

        Arguments:
            campaign (obj): The Campaign instance to update, with a valid "id".
        Return:
            A Campaign instance as the updated one.
        """
        assert campaign and isinstance(campaign, Campaign)
        campaign_id = campaign['id']
        assert campaign_id
        res = self.request_post('/campaign/{0}'.format(campaign_id), campaign)
        return Campaign.from_dict(res)

    def del_campaign(self, campaign):
        """ Delete a campaign.

        Arguments:
            campaign (obj or int): The Campaign instance or campaign id to delete.
        Return:
            The campaign id deleted. It must be the same as the one in argument.
        """
        assert campaign
        campaign_id = campaign['id'] if isinstance(campaign, Campaign) else int(campaign)
        assert campaign_id
        res = self.request_del('/campaign/{0}'.format(campaign_id))
        assert res == campaign_id
        return res

    # User filter #
    def get_user_filters(self, group):
        """ Get the user filters in a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to get.
        Return:
            A list of UserFilter instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_get('/group/{0}/userfilter'.format(group_id))
        return [UserFilter.from_dict(x) for x in res]

    def add_user_filter(self, group, user_filter, file_path):
        """ Add a new user filter data.

        The user_filter argument is setting only, and it requires a file containing the user list data.
        The file format is standard CSV, and include at least a "uid" column as the user/customer id list.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to add user filter.
            user_filter (obj): The UserFilter instance to add.
            file_path (str): The file path of the user list to upload.
        Return:
            A UserFilter instance as the added one (with fields filled by server, e.g. createTime).
        """
        assert group and user_filter and isinstance(user_filter, UserFilter)
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_upload('/group/{0}/userfilter'.format(group_id), user_filter, file_path)
        return UserFilter.from_dict(res)

    def update_user_filter(self, user_filter, file_path):
        """ Update an user filter data.

        The user_filter argument is setting only, and it requires a file containing the user list data.
        The file format is standard CSV, and include at least a "uid" column as the user/customer id list.

        Arguments:
            user_filter (obj): The UserFilter instance to update, with valid "id".
            file_path (str): The file path of the user list to upload.
        Return:
            A UserFilter instance as updated.
        """
        assert user_filter and isinstance(user_filter, UserFilter)
        user_filter_id = user_filter['id']
        assert user_filter_id
        res = self.request_upload('/userfilter/{0}'.format(user_filter_id), user_filter, file_path)
        return UserFilter.from_dict(res)

    def del_user_filter(self, user_filter):
        """ Delete an user filter.

        Arguments:
            user_filter (obj or int): The UserFilter instance or a user filter id to delete.
        Return:
            The user filter id deleted. It must be the same as the one in argument.
        """
        assert user_filter
        user_filter_id = user_filter['id'] if isinstance(user_filter, UserFilter) else int(user_filter)
        assert user_filter_id
        res = self.request_del('/userfilter/{0}'.format(user_filter_id))
        assert res == user_filter_id
        return res

    # Layout #
    def get_layout(self, logic):
        """ Get layout setting of a recommendation logic.

        If there is no layout saved for the given logic, it raises a 404 error. Client should handle the error.
        In this case, it is possible to create a new Layout instance with proper default value before further processing.

        Arguments:
            logic (obj or int): The Logic instance or a logic id to get layout.
        Return:
            A Layout instance.
        """
        assert logic
        logic_id = logic['id'] if isinstance(logic, Logic) else int(logic)
        res = self.request_get('/logic/{0}/layout'.format(logic_id))
        return Layout.from_dict(res)

    def update_layout(self, logic, layout):
        """ Add or update layout setting of a recommendation logic.

        Arguments:
            logic (obj or int): The Logic instance or a logic id to add/update layout.
            layout (obj): The Layout instance to add/update.
        Return:
            A Layout instance as added or updated.
        """
        assert logic and layout and isinstance(layout, Layout)
        logic_id = logic['id'] if isinstance(logic, Logic) else int(logic)
        res = self.request_post('/logic/{0}/layout'.format(logic_id), layout)
        return Layout.from_dict(res)

    def get_alg_trainings(self, group):
        """ Get algorithm training results in a group.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to get.
        Return:
            A list of AlgTraing instances.
        """
        assert group
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_get('/group/{0}/algtraining'.format(group_id))
        return [AlgTraining.from_dict(x) for x in res]

    def add_alg_training(self, group, alg_training):
        """ Submit a algorithm training.

        Arguments:
            group (obj or int): The emc.Group instance, ERGroup instance, or group id to submit alg training.
            alg_training (obj): The AlgTraining instance to submit.
        Return:
            A AlgTraining instance as the submitted one (with fields filled by server, e.g. createTime).            
        """
        assert group and alg_training and isinstance(alg_training, AlgTraining)
        group_id = group['id'] if isinstance(group, Group) or isinstance(group, ERGroup) else int(group)
        res = self.request_post('/group/{0}/algtraining'.format(group_id), alg_training)
        return AlgTraining.from_dict(res)



from setuptools import setup, find_packages

import etunexus

setup(
    name='etunexus',
    version=etunexus.__version__,
    description='Etu Nexus API',
    long_description=open("README.rst").read(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ],
    url=etunexus.__url__,
    author=etunexus.__author__,
    author_email=etunexus.__author_email__,
    license=etunexus.__license__,
    packages=find_packages(),
    install_requires=[
        'MultipartPostHandler'
    ],
    zip_safe=False)